# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest
from time import sleep

import mock

from copy import deepcopy
from datetime import datetime
from typing import Optional
from dateutil.tz import gettz
from mock import Mock
from mock.mock import call
from neon_utils.user_utils import get_user_prefs
from neon_utils.language_utils import SupportedLanguages
from ovos_bus_client import Message

from neon_minerva.tests.skill_unit_test_base import SkillTestCase


class TestSkill(SkillTestCase):
    test_message = Message("test", {}, {"neon_in_request": True})
    default_config = deepcopy(get_user_prefs())
    default_config['location']['country'] = "United States"

    @classmethod
    @mock.patch('neon_utils.language_utils.get_supported_languages')
    def setUpClass(cls, get_langs) -> None:
        get_langs.return_value = SupportedLanguages({'en'}, {'en'}, {'en'})
        SkillTestCase.setUpClass()

    def setUp(self):
        SkillTestCase.setUp(self)
        self.user_config = deepcopy(self.default_config)

    def test_00_skill_init(self):
        # Test any parameters expected to be set in init or initialize methods
        from neon_utils.skills.neon_skill import NeonSkill

        self.assertIsInstance(self.skill, NeonSkill)

    def test_stt_languages(self):
        real_languages = self.skill._languages
        # Languages Specified
        self.skill._languages = SupportedLanguages({'en', 'es', 'uk'},
                                                   {}, {'en', 'uk', 'pl'})
        self.assertEqual(self.skill.stt_languages, {'en', 'uk'})

        # Languages not Specified
        self.skill._languages = SupportedLanguages({}, {}, {'en', 'uk', 'pl'})
        self.assertIsNone(self.skill.stt_languages)

        self.skill._languages = real_languages

    def test_tts_languages(self):
        real_languages = self.skill._languages
        # Languages Specified
        self.skill._languages = SupportedLanguages({}, {'en', 'es', 'uk'},
                                                   {'en', 'uk', 'pl'})
        self.assertEqual(self.skill.tts_languages, {'en', 'uk'})

        # Languages not Specified
        self.skill._languages = SupportedLanguages({}, {'en', 'es', 'uk'}, {})
        self.assertIsNone(self.skill.tts_languages)

        self.skill._languages = real_languages

    def test_location_update(self):
        # TODO: Test ipgeo update at init
        pass

    def test_handle_unit_change(self):
        test_profile = self.user_config
        test_profile["user"]["username"] = "test_user"
        test_profile["units"]["measure"] = "imperial"
        test_message = Message("test", {"imperial": "imperial"},
                               {"username": "test_user",
                                "user_profiles": [test_profile]})
        # imperial -> imperial
        self.skill.handle_unit_change(test_message)
        self.skill.speak_dialog.assert_called_once_with("units_already_set",
                                                        {"unit": "imperial"},
                                                        private=True)
        self.assertEqual(
            test_message.context["user_profiles"][0]["units"]["measure"],
            "imperial")
        # imperial -> metric
        test_message.data = {"metric": "metric"}
        self.skill.handle_unit_change(test_message)
        self.skill.speak_dialog.assert_called_with("units_changed",
                                                   {"unit": "metric"},
                                                   private=True)
        self.assertEqual(
            test_message.context["user_profiles"][0]["units"]["measure"],
            "metric")
        # metric -> metric
        self.skill.handle_unit_change(test_message)
        self.skill.speak_dialog.assert_called_with("units_already_set",
                                                   {"unit": "metric"},
                                                   private=True)
        self.assertEqual(
            test_message.context["user_profiles"][0]["units"]["measure"],
            "metric")
        # metric -> imperial
        test_message.data = {"imperial": "imperial"}
        self.skill.handle_unit_change(test_message)
        self.skill.speak_dialog.assert_called_with("units_changed",
                                                   {"unit": "imperial"},
                                                   private=True)
        self.assertEqual(
            test_message.context["user_profiles"][0]["units"]["measure"],
            "imperial")

    def test_handle_time_format_change(self):
        test_profile = self.user_config
        test_profile["user"]["username"] = "test_user"
        test_profile["units"]["time"] = 12
        test_message = Message("test", {"half": "12 hour"},
                               {"username": "test_user",
                                "user_profiles": [test_profile]})
        # 12 -> 12
        self.skill.handle_time_format_change(test_message)
        self.skill.speak_dialog.assert_called_once_with(
            "time_format_already_set", {"scale": "12"}, private=True)
        self.assertEqual(
            test_message.context["user_profiles"][0]["units"]["time"], 12)
        # 12 -> 24
        test_message.data = {"full": "24 hour"}
        self.skill.handle_time_format_change(test_message)
        self.skill.speak_dialog.assert_called_with("time_format_changed",
                                                   {"scale": "24"},
                                                   private=True)
        self.assertEqual(
            test_message.context["user_profiles"][0]["units"]["time"], 24)
        # 24 -> 24
        self.skill.handle_time_format_change(test_message)
        self.skill.speak_dialog.assert_called_with("time_format_already_set",
                                                   {"scale": "24"},
                                                   private=True)
        self.assertEqual(
            test_message.context["user_profiles"][0]["units"]["time"], 24)
        # 24 -> 12
        test_message.data = {"half": "half scale"}
        self.skill.handle_time_format_change(test_message)
        self.skill.speak_dialog.assert_called_with("time_format_changed",
                                                   {"scale": "12"},
                                                   private=True)
        self.assertEqual(
            test_message.context["user_profiles"][0]["units"]["time"], 12)

    def test_handle_date_format_change(self):
        test_profile = self.user_config
        test_profile["user"]["username"] = "test_user"
        test_profile["units"]["date"] = "MDY"
        test_message = Message("test", {"mdy": "month day year"},
                               {"username": "test_user",
                                "user_profiles": [test_profile]})
        # MDY -> MDY
        self.skill.handle_date_format_change(test_message)
        self.skill.speak_dialog.assert_called_once_with(
            "date_format_already_set", {"format": "month day year"},
            private=True)
        self.assertEqual(
            test_message.context["user_profiles"][0]["units"]["date"], "MDY")
        # MDY -> YMD
        test_message.data = {"ymd": "year month day"}
        self.skill.handle_date_format_change(test_message)
        self.skill.speak_dialog.assert_called_with("date_format_changed",
                                                   {"format": "year month day"},
                                                   private=True)
        self.assertEqual(
            test_message.context["user_profiles"][0]["units"]["date"], "YMD")
        # YMD -> YMD
        self.skill.handle_date_format_change(test_message)
        self.skill.speak_dialog.assert_called_with("date_format_already_set",
                                                   {"format": "year month day"},
                                                   private=True)
        self.assertEqual(
            test_message.context["user_profiles"][0]["units"]["date"], "YMD")
        # YMD -> DMY
        test_message.data = {"dmy": "day month year"}
        self.skill.handle_date_format_change(test_message)
        self.skill.speak_dialog.assert_called_with("date_format_changed",
                                                   {"format": "day month year"},
                                                   private=True)
        self.assertEqual(
            test_message.context["user_profiles"][0]["units"]["date"], "DMY")

    def test_handle_speech_hesitation(self):
        test_profile = self.user_config
        test_profile["user"]["username"] = "test_user"
        test_profile["response_mode"]["hesitation"] = True
        test_message = Message("test", {"permit": "enable"},
                               {"username": "test_user",
                                "user_profiles": [test_profile]})
        # True -> True
        self.skill.handle_speak_hesitation(test_message)
        self.skill.speak_dialog.assert_called_once_with("hesitation_enabled",
                                                        private=True)
        self.assertTrue(test_message.context["user_profiles"][0]
                        ["response_mode"]["hesitation"])
        # True -> False
        test_message.data = {"deny": "disable"}
        self.skill.handle_speak_hesitation(test_message)
        self.skill.speak_dialog.assert_called_with("hesitation_disabled",
                                                   private=True)
        self.assertFalse(test_message.context["user_profiles"][0]
                         ["response_mode"]["hesitation"])
        # False -> False
        self.skill.handle_speak_hesitation(test_message)
        self.skill.speak_dialog.assert_called_with("hesitation_disabled",
                                                   private=True)
        self.assertFalse(test_message.context["user_profiles"][0]
                         ["response_mode"]["hesitation"])
        # False -> True
        test_message.data = {"permit": "enable"}
        self.skill.handle_speak_hesitation(test_message)
        self.skill.speak_dialog.assert_called_with("hesitation_enabled",
                                                   private=True)
        self.assertTrue(test_message.context["user_profiles"][0]
                        ["response_mode"]["hesitation"])

    def test_handle_transcription_retention(self):
        test_profile = self.user_config
        test_profile["user"]["username"] = "test_user"
        test_profile["privacy"]["save_audio"] = True
        test_profile["privacy"]["save_text"] = True
        test_message = Message("test", {"permit": "enable",
                                        "audio": "recordings"},
                               {"username": "test_user",
                                "user_profiles": [test_profile]})
        # Audio Recordings
        # Enable -> Enable
        self.skill.handle_transcription_retention(test_message)
        self.skill.speak_dialog.assert_called_with("transcription_already_set",
                                                   {"transcription": "audio",
                                                    "enabled": "enabled"},
                                                   private=True)
        self.assertTrue(test_message.context["user_profiles"][0]
                        ["privacy"]["save_audio"])
        # Enable -> Disable
        test_message.data = {"deny": "disable", "audio": "recording"}
        self.skill.handle_transcription_retention(test_message)
        self.skill.speak_dialog.assert_called_with("transcription_changed",
                                                   {"transcription": "audio",
                                                    "enabled": "disabled"},
                                                   private=True)
        self.assertFalse(test_message.context["user_profiles"][0]
                         ["privacy"]["save_audio"])
        # Disable -> Disable
        self.skill.handle_transcription_retention(test_message)
        self.skill.speak_dialog.assert_called_with("transcription_already_set",
                                                   {"transcription": "audio",
                                                    "enabled": "disabled"},
                                                   private=True)
        self.assertFalse(test_message.context["user_profiles"][0]
                         ["privacy"]["save_audio"])
        # Disable -> Enable
        test_message.data = {"permit": "enable", "audio": "recording"}
        self.skill.handle_transcription_retention(test_message)
        self.skill.speak_dialog.assert_called_with("transcription_changed",
                                                   {"transcription": "audio",
                                                    "enabled": "enabled"},
                                                   private=True)
        self.assertTrue(test_message.context["user_profiles"][0]
                        ["privacy"]["save_audio"])
        # Text Transcriptions
        # Enable -> Disable
        test_message.data = {"deny": "disable", "text": "transcription"}
        self.skill.handle_transcription_retention(test_message)
        self.skill.speak_dialog.assert_called_with("transcription_changed",
                                                   {"transcription": "text",
                                                    "enabled": "disabled"},
                                                   private=True)
        self.assertFalse(test_message.context["user_profiles"][0]
                         ["privacy"]["save_text"])
        # Disable -> Disable
        self.skill.handle_transcription_retention(test_message)
        self.skill.speak_dialog.assert_called_with("transcription_already_set",
                                                   {"transcription": "text",
                                                    "enabled": "disabled"},
                                                   private=True)
        self.assertFalse(test_message.context["user_profiles"][0]
                         ["privacy"]["save_text"])
        # Disable -> Enable
        test_message.data = {"permit": "allow", "text": "transcription"}
        self.skill.handle_transcription_retention(test_message)
        self.skill.speak_dialog.assert_called_with("transcription_changed",
                                                   {"transcription": "text",
                                                    "enabled": "enabled"},
                                                   private=True)
        self.assertTrue(test_message.context["user_profiles"][0]
                        ["privacy"]["save_text"])
        # Enable -> Enable
        self.skill.handle_transcription_retention(test_message)
        self.skill.speak_dialog.assert_called_with("transcription_already_set",
                                                   {"transcription": "text",
                                                    "enabled": "enabled"},
                                                   private=True)
        self.assertTrue(test_message.context["user_profiles"][0]
                        ["privacy"]["save_text"])

    def test_handle_speak_speed(self):
        test_profile = self.user_config
        test_profile["user"]["username"] = "test_user"
        test_message = Message("test", {"faster": "faster"},
                               {"username": "test_user",
                                "user_profiles": [test_profile]})
        self.assertEqual(test_profile["speech"]["speed_multiplier"], 1.0)

        # Speak faster
        self.skill.handle_speech_speed(test_message)
        self.skill.speak_dialog.assert_called_once_with("speech_speed_faster",
                                                        private=True)
        self.assertGreater(test_message.context["user_profiles"][0]
                           ["speech"]["speed_multiplier"], 1.0)
        # Speak max speed
        test_message.context["user_profiles"][0]["speech"][
            "speed_multiplier"] = self.skill.MAX_SPEECH_SPEED
        self.skill.handle_speech_speed(test_message)
        self.skill.speak_dialog.assert_called_with("speech_speed_limit",
                                                   {"limit": "faster"},
                                                   private=True)
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["speech"]["speed_multiplier"],
                         self.skill.MAX_SPEECH_SPEED)
        # Speak normally
        test_message.data = {"normally": "normally"}
        self.skill.handle_speech_speed(test_message)
        self.skill.speak_dialog.assert_called_with("speech_speed_normal",
                                                   private=True)
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["speech"]["speed_multiplier"], 1.0)
        # Speak slower
        test_message.data = {"slower": "slower"}
        self.skill.handle_speech_speed(test_message)
        self.skill.speak_dialog.assert_called_with("speech_speed_slower",
                                                   private=True)
        self.assertLess(test_message.context["user_profiles"][0]
                        ["speech"]["speed_multiplier"], 1.0)
        # Speak min speed
        test_message.context["user_profiles"][0]["speech"][
            "speed_multiplier"] = self.skill.MIN_SPEECH_SPEED
        self.skill.handle_speech_speed(test_message)
        self.skill.speak_dialog.assert_called_with("speech_speed_limit",
                                                   {"limit": "slower"},
                                                   private=True)
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["speech"]["speed_multiplier"],
                         self.skill.MIN_SPEECH_SPEED)

    def test_handle_change_location_timezone(self):
        real_ask_yesno = self.skill.ask_yesno
        self.skill.ask_yesno = Mock(return_value="no")

        test_profile = self.user_config
        test_profile["user"]["username"] = "test_user"
        test_message: Optional[Message] = None

        def _init_test_message(voc, location):
            nonlocal test_message
            test_message = Message("test", {voc: voc,
                                            "rx_place": location},
                                   {"username": "test_user",
                                    "user_profiles": [test_profile]})

        # Change location same tz
        _init_test_message("location", "new york")
        sleep(1)
        self.skill.handle_change_location_timezone(test_message)
        self.skill.ask_yesno.assert_called_once_with(
            "also_change_location_tz", {"type": "timezone", "new": "new york"})
        self.skill.speak_dialog.assert_called_once_with(
            "change_location_tz", {"type": "location", "location": "New York"},
            private=True)
        profile = test_message.context["user_profiles"][0]
        unchanged = ("tz", "utc", "country")
        for setting in unchanged:
            self.assertEqual(profile["location"][setting],
                             test_profile["location"][setting])
        self.assertEqual(profile["location"]["city"], "New York")
        self.assertEqual(profile["location"]["state"], "New York")
        self.assertAlmostEqual(profile["location"]["lat"], 40.7127281, 5)
        self.assertAlmostEqual(profile["location"]["lng"], -74.0060152, 5)
        self.skill.ask_yesno.reset_mock()
        self.skill.speak_dialog.reset_mock()

        # Change tz same location
        _init_test_message("timezone", "phoenix")
        sleep(1)
        self.skill.handle_change_location_timezone(test_message)
        self.skill.ask_yesno.assert_called_once_with(
            "also_change_location_tz", {"type": "location", "new": "phoenix"})
        self.skill.speak_dialog.assert_called_once_with(
            "change_location_tz", {"type": "timezone", "location": "UTC -7.0"},
            private=True)
        profile = test_message.context["user_profiles"][0]
        unchanged = ("city", "state", "country", "lat", "lng")
        for setting in unchanged:
            self.assertEqual(profile["location"][setting],
                             test_profile["location"][setting])
        self.assertEqual(profile["location"]["tz"], "America/Phoenix")
        self.assertEqual(profile["location"]["utc"], -7.0)
        self.skill.speak_dialog.reset_mock()
        self.skill.ask_yesno = Mock(return_value="yes")

        # Change location and tz
        _init_test_message("location", "honolulu")
        sleep(1)
        self.skill.handle_change_location_timezone(test_message)
        self.skill.ask_yesno.assert_called_once_with(
            "also_change_location_tz", {"type": "timezone", "new": "honolulu"})
        self.skill.speak_dialog.assert_has_calls((
            call("change_location_tz",
                 {"type": "location", "location": "Honolulu"}, private=True),
            call("change_location_tz",
                 {"type": "timezone", "location": "UTC -10.0"}, private=True)),
            True)

        profile = test_message.context["user_profiles"][0]
        unchanged = ("country",)
        for setting in unchanged:
            self.assertEqual(profile["location"][setting],
                             test_profile["location"][setting])
        self.assertEqual(profile["location"]["city"], "Honolulu")
        self.assertEqual(profile["location"]["state"], "Hawaii")
        self.assertAlmostEqual(profile["location"]["lat"], 21.2890997, 0)
        self.assertAlmostEqual(profile["location"]["lng"], -157.717299, 0)
        self.assertEqual(profile["location"]["tz"], "Pacific/Honolulu")
        self.assertEqual(profile["location"]["utc"], -10.0)

        self.skill.speak_dialog.reset_mock()
        self.skill.ask_yesno.reset_mock()

        # Change tz and location
        _init_test_message("timezone", "phoenix")
        sleep(1)
        self.skill.handle_change_location_timezone(test_message)
        self.skill.ask_yesno.assert_called_once_with(
            "also_change_location_tz", {"type": "location", "new": "phoenix"})

        self.skill.speak_dialog.assert_has_calls((
            call("change_location_tz",
                 {"type": "location", "location": "Phoenix"}, private=True),
            call("change_location_tz",
                 {"type": "timezone", "location": "UTC -7.0"}, private=True)),
            True)

        profile = test_message.context["user_profiles"][0]
        unchanged = ("country",)
        for setting in unchanged:
            self.assertEqual(profile["location"][setting],
                             test_profile["location"][setting])

        self.assertEqual(profile["location"]["city"], "Phoenix")
        self.assertEqual(profile["location"]["state"], "Arizona")
        self.assertAlmostEqual(profile["location"]["lat"], 33.4484367, 5)
        self.assertAlmostEqual(profile["location"]["lng"], -112.074141, 5)
        self.assertEqual(profile["location"]["tz"], "America/Phoenix")
        self.assertEqual(profile["location"]["utc"], -7.0)

        self.skill.ask_yesno = real_ask_yesno

    def test_handle_change_dialog_option(self):
        test_profile = self.user_config
        test_profile["user"]["username"] = "test_user"
        test_message = Message("test", {"limited": "limited"},
                               {"username": "test_user",
                                "user_profiles": [test_profile]})
        test_profile["response_mode"]["limit_dialog"] = False

        # random -> limited
        self.skill.handle_change_dialog_mode(test_message)
        self.assertTrue(test_message.context["user_profiles"][0]
                        ["response_mode"]["limit_dialog"])
        self.skill.speak_dialog.assert_called_with("dialog_mode_changed",
                                                   {"response": "limited"},
                                                   private=True)
        # limited -> limited
        self.skill.handle_change_dialog_mode(test_message)
        self.assertTrue(test_message.context["user_profiles"][0]
                        ["response_mode"]["limit_dialog"])
        self.skill.speak_dialog.assert_called_with("dialog_mode_already_set",
                                                   {"response": "limited"},
                                                   private=True)
        test_message.data = {"random": "standard"}
        # limited -> random
        self.skill.handle_change_dialog_mode(test_message)
        self.assertFalse(test_message.context["user_profiles"][0]
                         ["response_mode"]["limit_dialog"])
        self.skill.speak_dialog.assert_called_with("dialog_mode_changed",
                                                   {"response": "random"},
                                                   private=True)
        # random -> random
        self.skill.handle_change_dialog_mode(test_message)
        self.assertFalse(test_message.context["user_profiles"][0]
                         ["response_mode"]["limit_dialog"])
        self.skill.speak_dialog.assert_called_with("dialog_mode_already_set",
                                                   {"response": "random"},
                                                   private=True)

    def test_handle_say_my_name(self):
        test_profile = self.user_config
        test_message = Message("test", {},
                               {"username": "test_user",
                                "user_profiles": [test_profile]})

        # No name set
        self.skill.handle_say_my_name(test_message)
        self.skill.speak_dialog.assert_called_with("name_not_known",
                                                   {"name_position": "name"},
                                                   private=True)

        test_message.context["user_profiles"][0]["user"]["username"] = \
            "test_user"
        # first name not known
        test_message.data['utterance'] = "tell me my first name"
        self.skill.handle_say_my_name(test_message)
        self.skill.speak_dialog.assert_called_with(
            "name_not_known", {"name_position": "first name"}, private=True)
        # middle name not known
        test_message.data['utterance'] = "tell me my middle name"
        self.skill.handle_say_my_name(test_message)
        self.skill.speak_dialog.assert_called_with(
            "name_not_known", {"name_position": "middle name"}, private=True)
        # last name not known
        test_message.data['utterance'] = "tell me my last name"
        self.skill.handle_say_my_name(test_message)
        self.skill.speak_dialog.assert_called_with(
            "name_not_known", {"name_position": "last name"}, private=True)
        # preferred name not known
        test_message.data['utterance'] = "tell me my nick name"
        self.skill.handle_say_my_name(test_message)
        self.skill.speak_dialog.assert_called_with(
            "name_not_known", {"name_position": "preferred name"}, private=True)

        # name username set
        test_message.data['utterance'] = "tell me my name"
        self.skill.handle_say_my_name(test_message)
        self.skill.speak_dialog.assert_called_with(
            "name_is", {"name_position": "name",
                        "name": "test_user"}, private=True)

        # name first name set
        test_message.context["user_profiles"][0]["user"]["first_name"] = \
            "First"
        self.skill.handle_say_my_name(test_message)
        self.skill.speak_dialog.assert_called_with(
            "name_is", {"name_position": "name",
                        "name": "First"}, private=True)

        # name preferred name set
        test_message.context["user_profiles"][0]["user"]["preferred_name"] = \
            "Preferred"
        self.skill.handle_say_my_name(test_message)
        self.skill.speak_dialog.assert_called_with(
            "name_is", {"name_position": "name",
                        "name": "Preferred"}, private=True)

        test_message.context["user_profiles"][0]["user"]["middle_name"] = \
            "Middle"
        test_message.context["user_profiles"][0]["user"]["last_name"] = \
            "Last"

        # request first name
        test_message.data['utterance'] = "tell me my first name"
        self.skill.handle_say_my_name(test_message)
        self.skill.speak_dialog.assert_called_with(
            "name_is", {"name_position": "first name",
                        "name": "First"}, private=True)
        # request middle name
        test_message.data['utterance'] = "tell me my middle name"
        self.skill.handle_say_my_name(test_message)
        self.skill.speak_dialog.assert_called_with(
            "name_is", {"name_position": "middle name",
                        "name": "Middle"}, private=True)
        # request last name
        test_message.data['utterance'] = "tell me my last name"
        self.skill.handle_say_my_name(test_message)
        self.skill.speak_dialog.assert_called_with(
            "name_is", {"name_position": "last name",
                        "name": "Last"}, private=True)

        # request preferred name
        test_message.data['utterance'] = "tell me my preferred name"
        self.skill.handle_say_my_name(test_message)
        self.skill.speak_dialog.assert_called_with(
            "name_is", {"name_position": "preferred name",
                        "name": "Preferred"}, private=True)

        # request username
        test_message.data['utterance'] = "tell me my username"
        self.skill.handle_say_my_name(test_message)
        self.skill.speak_dialog.assert_called_with(
            "name_is", {"name_position": "username",
                        "name": "test_user"}, private=True)

        # Username not supported
        test_message.data['utterance'] = "tell me my username"
        test_message.context['username'] = 'local'
        test_message.context['user_profiles'][0]['user']['username'] = 'local'
        self.skill.handle_say_my_name(test_message)
        self.skill.speak_dialog.assert_called_with("name_no_username",
                                                   private=True)

    def test_handle_say_my_email(self):
        test_profile = self.user_config
        test_profile["user"]["username"] = "test_user"
        test_message = Message("test", {},
                               {"username": "test_user",
                                "user_profiles": [test_profile]})
        self.assertFalse(test_profile["user"]["email"])
        self.skill.handle_say_my_email(test_message)
        self.skill.speak_dialog.assert_called_with("email_not_known",
                                                   private=True)
        test_message.context["user_profiles"][0]["user"]["email"] = \
            "test@neon.ai"
        self.skill.handle_say_my_email(test_message)
        self.skill.speak_dialog.assert_called_with(
            "email_is", {"email": "test at neon dot ai"}, private=True)

    @mock.patch('neon_utils.net_utils.check_online')
    def test_handle_say_my_location(self, check_online):
        test_profile = self.user_config
        test_profile["user"]["username"] = "test_user"
        test_profile['location'] = {'lat': '',
                                    'lng': '',
                                    'city': '',
                                    'state': '',
                                    'country': ''}
        test_message = Message("test", {},
                               {"username": "test_user",
                                "user_profiles": [test_profile]})

        # No location, offline
        check_online.return_value = False
        self.skill.handle_say_my_location(test_message)
        self.skill.speak_dialog.assert_called_with("location_unknown_offline",
                                                   private=True)
        # No location, online
        check_online.return_value = True
        self.skill.handle_say_my_location(test_message)
        self.skill.speak_dialog.assert_called_with("location_unknown_online",
                                                   private=True)
        # Valid city, state, country
        test_message.context['user_profiles'][0]['location']['city'] = "Renton"
        test_message.context['user_profiles'][0]['location']['state'] = \
            "Washington"
        test_message.context['user_profiles'][0]['location']['country'] = \
            "United States"
        self.skill.handle_say_my_location(test_message)
        self.skill.speak_dialog.assert_called_with(
            "location_is", {"location": "Renton, Washington"}, private=True)

        # Valid city, country
        test_message.context['user_profiles'][0]['location']['city'] = "Kyiv"
        test_message.context['user_profiles'][0]['location']['state'] = ""
        test_message.context['user_profiles'][0]['location']['country'] = \
            "Ukraine"
        self.skill.handle_say_my_location(test_message)
        self.skill.speak_dialog.assert_called_with(
            "location_is", {"location": "Kyiv, Ukraine"}, private=True)

    def test_handle_set_my_birthday(self):
        test_profile = self.user_config
        test_profile["user"]["username"] = "test_user"
        test_profile['location']['tz'] = 'America/Los_Angeles'

        test_message = Message("test", {"utterance": "my birthday is today"},
                               {"username": "test_user",
                                "user_profiles": [test_profile]})
        self.skill.handle_set_my_birthday(test_message)
        today = datetime.now(gettz(test_profile["location"]["tz"]))
        self.skill.speak_dialog.assert_has_calls((
            call("birthday_confirmed", {"birthday": today.strftime("%B %-d")},
                 private=True),
            call("happy_birthday", private=True)
        ), True)

        test_message.data['utterance'] = "my birthday is september 9"
        self.skill.handle_set_my_birthday(test_message)
        self.skill.speak_dialog.assert_called_with("birthday_confirmed",
                                                   {"birthday": "September 9"},
                                                   private=True)

    def test_handle_say_my_birthday(self):
        test_profile = self.user_config
        test_profile["user"]["username"] = "test_user"
        test_profile['location']['tz'] = 'America/Los_Angeles'
        # Not Set
        test_message = Message("test", {"utterance": "my birthday is today"},
                               {"username": "test_user",
                                "user_profiles": [test_profile]})
        self.skill.handle_say_my_birthday(test_message)
        self.skill.speak_dialog.assert_called_with("birthday_not_known",
                                                   private=True)

        # Set
        test_profile["user"]["dob"] = "2000/01/01"
        test_message = Message("test", {"utterance": "my birthday is today"},
                               {"username": "test_user",
                                "user_profiles": [test_profile]})
        self.skill.handle_say_my_birthday(test_message)
        self.skill.speak_dialog.assert_any_call("birthday_is",
                                                {"birthday": "January 1"},
                                                private=True)

        # Today
        now_time = datetime.now(gettz(test_profile['location']['tz']))
        test_profile["user"]["dob"] = now_time.strftime("%Y/%m/%d")
        test_message = Message("test", {"utterance": "my birthday is today"},
                               {"username": "test_user",
                                "user_profiles": [test_profile]})
        self.skill.handle_say_my_birthday(test_message)
        self.skill.speak_dialog.assert_any_call("happy_birthday", private=True)

    def test_handle_set_my_email(self):
        real_get_gui_input = self.skill.get_gui_input
        self.skill.get_gui_input = Mock(return_value=None)
        real_ask_yesno = self.skill.ask_yesno
        self.skill.ask_yesno = Mock(return_value="no")
        test_profile = self.user_config
        test_profile["user"]["username"] = "test_user"
        test_message = Message("test", {},
                               {"username": "test_user",
                                "user_profiles": [test_profile]})

        def _check_not_confirmed(msg):
            self.skill.handle_set_my_email(msg)
            self.skill.ask_yesno.assert_called_once_with(
                "email_confirmation", {"email": "test at neon dot ai"})
            self.skill.speak_dialog.assert_called_once_with(
                "email_not_confirmed", private=True)
            self.skill.ask_yesno.reset_mock()
            self.skill.speak_dialog.reset_mock()

        # Typed input
        test_message.data["utterance"] = "my email address is test@neon.ai"
        test_message.data["rx_setting"] = "test@neon . "
        _check_not_confirmed(test_message)

        # Valid parse
        test_message.data["utterance"] = "my email address is test@neon.ai"
        test_message.data["rx_setting"] = "test@neon.ai"
        _check_not_confirmed(test_message)

        # Valid parse
        test_message.data["utterance"] = "my email address is test@neon. ai"
        test_message.data["rx_setting"] = "test@neon"
        _check_not_confirmed(test_message)

        # Spoken input recognized domain
        test_message.data["utterance"] = "my email address is test at neon.ai"
        test_message.data["rx_setting"] = "test at neon ."
        _check_not_confirmed(test_message)

        # Spoken input unrecognized domain
        test_message.data["utterance"] = "my email is test at neon dot ai"
        test_message.data["rx_setting"] = "test at neon dot ai"
        _check_not_confirmed(test_message)

        # Invalid address
        test_message.data["utterance"] = "my email address is test at neon ai"
        test_message.data["rx_setting"] = "test at neon ai"
        self.skill.handle_set_my_email(test_message)
        self.skill.speak_dialog.assert_called_once_with(
            "email_set_error", private=True)
        self.skill.speak_dialog.reset_mock()

        # Set Email Confirmed
        test_message.data["utterance"] = "my email is test at neon dot ai"
        test_message.data["rx_setting"] = "test at neon dot ai"
        self.skill.ask_yesno = Mock(return_value="yes")
        self.skill.handle_set_my_email(test_message)
        self.skill.ask_yesno.assert_called_with("email_confirmation",
                                                {"email": "test at neon dot ai"})
        self.skill.speak_dialog.assert_called_with("email_set",
                                                   {"email": "test at neon dot ai"},
                                                   private=True)
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["email"], "test@neon.ai")
        # Set Email No Change
        self.skill.handle_set_my_email(test_message)
        self.skill.speak_dialog.assert_called_with("email_already_set_same",
                                                   {"email": "test at neon dot ai"},
                                                   private=True)
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["email"], "test@neon.ai")
        # Change Email Not Confirmed
        self.skill.ask_yesno = Mock(return_value="no")
        test_message.data["utterance"] = "my email is demo at neon dot ai"
        test_message.data["rx_setting"] = "demo at neon dot ai"
        self.skill.handle_set_my_email(test_message)
        self.skill.ask_yesno.assert_called_with("email_overwrite",
                                                {"old": "test at neon dot ai",
                                                 "new": "demo at neon dot ai"})
        self.skill.speak_dialog.assert_called_with("email_not_changed",
                                                   {"email": "test at neon dot ai"},
                                                   private=True)
        # Change Email Confirmed
        self.skill.ask_yesno = Mock(return_value="yes")
        self.skill.handle_set_my_email(test_message)
        self.skill.ask_yesno.assert_called_with("email_overwrite",
                                                {"old": "test at neon dot ai",
                                                 "new": "demo at neon dot ai"})
        self.skill.speak_dialog.assert_called_with("email_set",
                                                   {"email": "demo at neon dot ai"},
                                                   private=True)
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["email"], "demo@neon.ai")

        self.skill.ask_yesno = real_ask_yesno
        self.skill.get_gui_input = real_get_gui_input

    def test_handle_set_my_name(self):
        test_profile = self.user_config
        test_profile["user"]["username"] = "test_user"
        test_message = Message("test", {},
                               {"username": "test_user",
                                "user_profiles": [test_profile]})

        # Set first name
        test_message.data["utterance"] = "my first name is daniel"
        test_message.data["rx_setting"] = "daniel"
        self.skill.handle_set_my_name(test_message)
        self.skill.speak_dialog.assert_called_with("name_set_part",
                                                   {"position": "first name",
                                                    "name": "Daniel"},
                                                   private=True)
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["first_name"], "Daniel")
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["full_name"], "Daniel")
        # Set middle name
        test_message.data["utterance"] = "my middle name is james"
        test_message.data["rx_setting"] = "james"
        self.skill.handle_set_my_name(test_message)
        self.skill.speak_dialog.assert_called_with("name_set_part",
                                                   {"position": "middle name",
                                                    "name": "James"},
                                                   private=True)
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["middle_name"], "James")
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["full_name"], "Daniel James")
        # Set last name
        test_message.data["utterance"] = "my last name is McKnight"
        test_message.data["rx_setting"] = "McKnight"
        self.skill.handle_set_my_name(test_message)
        self.skill.speak_dialog.assert_called_with("name_set_part",
                                                   {"position": "last name",
                                                    "name": "Mcknight"},
                                                   private=True)
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["last_name"], "Mcknight")
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["full_name"], "Daniel James Mcknight")
        # Set preferred name
        test_message.data["utterance"] = "my preferred name is Dan"
        test_message.data["rx_setting"] = "Dan"
        self.skill.handle_set_my_name(test_message)
        self.skill.speak_dialog.assert_called_with(
            "name_set_part",
            {"position": "preferred name", "name": "Dan"}, private=True)
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["preferred_name"], "Dan")
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["full_name"], "Daniel James Mcknight")
        # Set preferred name unchanged
        self.skill.handle_set_my_name(test_message)
        self.skill.speak_dialog.assert_called_with(
            "name_not_changed",
            {"position": "preferred name", "name": "Dan"}, private=True)
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["preferred_name"], "Dan")
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["full_name"], "Daniel James Mcknight")
        # Set full name no change
        test_message.data["utterance"] = "my name is Daniel James McKnight"
        test_message.data["rx_setting"] = "Daniel James McKnight"
        self.skill.handle_set_my_name(test_message)
        self.skill.speak_dialog.assert_called_with(
            "name_not_changed", {"position": "name",
                                 "name": "Daniel James Mcknight"})
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["full_name"], "Daniel James Mcknight")
        # Set full name 1 first changed
        test_message.data["utterance"] = "my name is test"
        test_message.data["rx_setting"] = "test"
        self.skill.handle_set_my_name(test_message)
        self.skill.speak_dialog.assert_called_with(
            "name_set_full", {"nick": "Dan", "name": "Test James Mcknight"},
            private=True)
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["full_name"], "Test James Mcknight")
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["first_name"], "Test")
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["preferred_name"], "Dan")
        # Set full name 2 last changed
        test_message.data["utterance"] = "my name is test this user"
        test_message.data["rx_setting"] = "test this user"
        self.skill.handle_set_my_name(test_message)
        self.skill.speak_dialog.assert_called_with(
            "name_set_full", {"nick": "Dan", "name": "Test This User"},
            private=True)
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["full_name"], "Test This User")
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["middle_name"], "This")
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["last_name"], "User")
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["preferred_name"], "Dan")
        # Set full name 4
        test_message.data["utterance"] = "my name is test this user again"
        test_message.data["rx_setting"] = "test this user again"
        self.skill.handle_set_my_name(test_message)
        self.skill.speak_dialog.assert_called_with(
            "name_set_full", {"nick": "Dan", "name": "Test This User Again"},
            private=True)
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["full_name"], "Test This User Again")
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["middle_name"], "This")
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["last_name"], "User Again")
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["preferred_name"], "Dan")

        real_ask_yesno = self.skill.ask_yesno
        self.skill.ask_yesno = Mock(return_value=False)

        # Test first name too long, unconfirmed
        test_message.data['utterance'] = ("my first name is some misfired "
                                          "intent match")
        test_message.data["rx_setting"] = "some misfired intent match"
        self.skill.handle_set_my_name(test_message)
        self.skill.ask_yesno.assert_called_with("name_confirm_change",
                                                {"position": "first name",
                                                 "name": test_message.data[
                                                     "rx_setting"].title()})
        self.skill.speak_dialog.assert_called_with("name_not_confirmed",
                                                   private=True)

        # Test full name too long, unconfirmed
        test_message.data['utterance'] = ("my name is some other misfired "
                                          "intent match")
        test_message.data["rx_setting"] = "some other misfired intent match"
        self.skill.handle_set_my_name(test_message)
        self.skill.ask_yesno.assert_called_with("name_confirm_change",
                                                {"position": "name",
                                                 "name": test_message.data[
                                                     "rx_setting"].title()})
        self.skill.speak_dialog.assert_called_with("name_not_confirmed",
                                                   private=True)

        self.skill.ask_yesno.return_value = True
        # Test full name too long, confirmed
        test_message.data['utterance'] = ("my name is José Eduardo Santos "
                                          "Tavares Melo Silva")
        test_message.data["rx_setting"] = "José Eduardo Santos Tavares Melo Silva"
        self.skill.handle_set_my_name(test_message)
        self.skill.ask_yesno.assert_called_with("name_confirm_change",
                                                {"position": "name",
                                                 "name": test_message.data[
                                                     "rx_setting"].title()})
        self.skill.speak_dialog.assert_called_with(
            "name_set_full",
            {"nick": test_message.context["user_profiles"][0]["user"]
             ["preferred_name"],
             "name": test_message.data["rx_setting"]}, private=True)
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["full_name"], test_message.data["rx_setting"])
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["first_name"], "José")
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["middle_name"], "Eduardo")
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["user"]["last_name"], "Santos Tavares Melo Silva")

        # Test last name too long, confirmed no change
        test_message.data['utterance'] = ("my last name is Santos "
                                          "Tavares Melo Silva")
        test_message.data["rx_setting"] = "Santos Tavares Melo Silva"
        self.skill.handle_set_my_name(test_message)
        self.skill.ask_yesno.assert_called_with("name_confirm_change",
                                                {"position": "last name",
                                                 "name": test_message.data[
                                                     "rx_setting"].title()})
        self.skill.speak_dialog.assert_called_with("name_not_changed",
                                                   {"position": "last name",
                                                    "name": test_message.data[
                                                        "rx_setting"]},
                                                   private=True)

        self.skill.ask_yesno = real_ask_yesno

    def test_handle_say_my_language_settings(self):
        test_profile = self.user_config
        test_profile["user"]["username"] = "test_user"
        test_profile["speech"]["tts_language"] = "en-us"
        test_profile["speech"]["tts_gender"] = "female"
        test_message = Message("test", {},
                               {"username": "test_user",
                                "user_profiles": [test_profile]})
        # Only one lang
        self.skill.handle_say_my_language_settings(test_message)
        self.skill.speak_dialog.assert_called_once_with(
            "language_setting", {"primary": "primary",
                                 "language": "American English",
                                 "gender": "female"}, private=True)
        # Two diff langs
        test_message.context["user_profiles"][0][
            "speech"]["secondary_tts_language"] = "uk-ua"
        self.skill.speak_dialog.reset_mock()
        self.skill.handle_say_my_language_settings(test_message)
        self.skill.speak_dialog.assert_has_calls((
            call("language_setting", {"primary": "primary",
                                      "language": "American English",
                                      "gender": "female"}, private=True),
            call("language_setting", {"primary": "secondary",
                                      "language": "Ukrainian",
                                      "gender": "male"}, private=True)
        ))
        # Two langs same gender
        test_message.context["user_profiles"][0][
            "speech"]["secondary_tts_gender"] = "female"
        self.skill.speak_dialog.reset_mock()
        self.skill.handle_say_my_language_settings(test_message)
        self.skill.speak_dialog.assert_has_calls((
            call("language_setting", {"primary": "primary",
                                      "language": "American English",
                                      "gender": "female"}, private=True),
            call("language_setting", {"primary": "secondary",
                                      "language": "Ukrainian",
                                      "gender": "female"}, private=True)
        ))
        # One lang diff gender
        test_message.context["user_profiles"][0][
            "speech"]["secondary_tts_gender"] = "male"
        test_message.context["user_profiles"][0][
            "speech"]["secondary_tts_language"] = "en-us"
        self.skill.speak_dialog.reset_mock()
        self.skill.handle_say_my_language_settings(test_message)
        self.skill.speak_dialog.assert_has_calls((
            call("language_setting", {"primary": "primary",
                                      "language": "American English",
                                      "gender": "female"}, private=True),
            call("language_setting", {"primary": "secondary",
                                      "language": "American English",
                                      "gender": "male"}, private=True)
        ))

    def test_handle_set_stt_language(self):
        real_ask_yesno = self.skill.ask_yesno
        real_supported_languages = self.skill._languages
        test_profile = self.user_config
        test_profile["user"]["username"] = "test_user"
        test_profile["speech"]["stt_language"] = "en-us"
        test_message = Message("test", {"rx_language": "something"},
                               {"username": "test_user",
                                "user_profiles": [test_profile]})
        # Invalid lang
        self.skill.handle_set_stt_language(test_message)
        self.skill.speak_dialog.assert_called_with("language_not_recognized",
                                                   {"lang": "something"},
                                                   private=True)
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["speech"]["stt_language"], "en-us")

        # Same lang
        test_message.data["rx_language"] = "english"
        self.skill.handle_set_stt_language(test_message)
        self.skill.speak_dialog.assert_called_with(
            "language_not_changed", {"io": "input",
                                     "lang": "American English"},
            private=True)

        # Change lang unsupported
        test_message.data["rx_language"] = "ukrainian"
        self.skill.handle_set_stt_language(test_message)
        self.skill.speak_dialog.assert_called_with("language_not_supported",
                                                   {"io": "understand",
                                                    "lang": "Ukrainian"},
                                                   private=True)

        # Mock supported STT langs
        self.skill._languages = SupportedLanguages({'uk'}, {}, {'uk'})

        # Change lang unconfirmed
        self.skill.ask_yesno = Mock(return_value=False)
        self.skill.handle_set_stt_language(test_message)
        self.skill.ask_yesno.assert_called_once_with(
            "language_change_confirmation", {"io": "input",
                                             "lang": "Ukrainian"})
        self.skill.speak_dialog.assert_called_with("language_not_confirmed",
                                                   private=True)
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["speech"]["stt_language"], "en-us")

        # Change lang confirmed
        self.skill.ask_yesno = Mock(return_value="yes")
        self.skill.handle_set_stt_language(test_message)
        self.skill.ask_yesno.assert_called_once_with(
            "language_change_confirmation", {"io": "input",
                                             "lang": "Ukrainian"})
        self.skill.speak_dialog.assert_called_with("language_set",
                                                   {"io": "input",
                                                    "lang": "Ukrainian"},
                                                   private=True)
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["speech"]["stt_language"], "uk-ua")

        self.skill.ask_yesno = real_ask_yesno
        self.skill._languages = real_supported_languages

    def test_handle_set_tts_language(self):
        real_supported_languages = self.skill._languages
        self.skill._languages = SupportedLanguages({}, {'it', 'ja', 'de', 'uk',
                                                        'pl', 'fr', 'es', 'bg',
                                                        'en'},
                                                   {'it', 'ja', 'de', 'uk',
                                                    'pl', 'fr', 'es',
                                                    'en'})

        test_profile = self.user_config
        test_profile["user"]["username"] = "test_user"
        test_message = Message("test", {},
                               {"username": "test_user",
                                "user_profiles": [test_profile]})
        # Change TTS language
        test_message.data = {"utterance": "Change my TTS language to spanish",
                             "rx_language": "spanish"}
        self.skill.handle_set_tts_language(test_message)
        self.skill.speak_dialog.assert_called_with(
            "language_set", {"io": "primary", "lang": "Spanish"}, private=True)
        self.assertEqual(test_message.context["user_profiles"][0]["speech"]
                         ["tts_language"], "es-es")
        self.assertEqual(test_message.context["user_profiles"][0]["speech"]
                         ["tts_gender"], test_profile["speech"]["tts_gender"])
        # Change Primary TTS language
        test_message.data = {"utterance": "Change my Primary TTS language "
                                          "to french",
                             "rx_language": "french"}
        self.skill.handle_set_tts_language(test_message)
        self.skill.speak_dialog.assert_called_with(
            "language_set", {"io": "primary", "lang": "French"}, private=True)
        self.assertEqual(test_message.context["user_profiles"][0]["speech"]
                         ["tts_language"], "fr-fr")
        self.assertEqual(test_message.context["user_profiles"][0]["speech"]
                         ["tts_gender"], test_profile["speech"]["tts_gender"])
        # Change Secondary TTS language
        test_message.data = {"utterance": "Change my secondary TTS language "
                                          "to female Polish"}
        self.skill.handle_set_tts_language(test_message)
        self.skill.speak_dialog.assert_called_with(
            "language_set", {"io": "secondary", "lang": "Polish"},
            private=True)
        self.assertEqual(test_message.context["user_profiles"][0]["speech"]
                         ["secondary_tts_language"], "pl-pl")
        self.assertEqual(test_message.context["user_profiles"][0]["speech"]
                         ["secondary_tts_gender"], "female")
        # Talk to me
        test_message.data = {"utterance": "Talk to me in Ukrainian",
                             "rx_language": "Ukrainian"}
        self.skill.handle_set_tts_language(test_message)
        self.skill.speak_dialog.assert_called_with(
            "language_set", {"io": "primary", "lang": "Ukrainian"},
            private=True)
        self.assertEqual(test_message.context["user_profiles"][0]["speech"]
                         ["tts_language"], "uk-ua")
        self.assertEqual(test_message.context["user_profiles"][0]["speech"]
                         ["tts_gender"], test_profile["speech"]["tts_gender"])
        # Change primary and secondary
        test_message.data = {"utterance": "Change my primary language to "
                                          "English and my secondary language "
                                          "to German"}
        self.skill.speak_dialog.reset_mock()
        self.skill.handle_set_tts_language(test_message)
        self.skill.speak_dialog.assert_has_calls((
            call("language_set",
                 {"io": "primary", "lang": "American English"}, private=True),
            call("language_set",
                 {"io": "secondary", "lang": "German"}, private=True)
        ))
        self.assertEqual(test_message.context["user_profiles"][0]["speech"]
                         ["tts_language"], "en-us")
        self.assertEqual(test_message.context["user_profiles"][0]["speech"]
                         ["secondary_tts_language"], "de-de")
        # Change primary language invalid and secondary language valid
        test_message.data = {"utterance": "Change my primary language to "
                                          "nothing and my second language "
                                          "to Italian"}
        self.skill.speak_dialog.reset_mock()
        self.skill.handle_set_tts_language(test_message)
        self.skill.speak_dialog.assert_has_calls((
            call("language_not_recognized",
                 {"lang": "nothing"}, private=True),
            call("language_set",
                 {"io": "secondary", "lang": "Italian"}, private=True)
        ))
        self.assertEqual(test_message.context["user_profiles"][0]["speech"]
                         ["tts_language"], "en-us")
        self.assertEqual(test_message.context["user_profiles"][0]["speech"]
                         ["secondary_tts_language"], "it-it")

        # Change primary language valid and second language invalid
        test_message.data = {"utterance": "Change my primary language to "
                                          "Japanese and my second language "
                                          "to something"}
        self.skill.speak_dialog.reset_mock()
        self.skill.handle_set_tts_language(test_message)
        self.skill.speak_dialog.assert_has_calls((
            call("language_set",
                 {"io": "primary", "lang": "Japanese"}, private=True),
            call("language_not_recognized",
                 {"lang": "something"}, private=True)

        ))
        self.assertEqual(test_message.context["user_profiles"][0]["speech"]
                         ["tts_language"], "ja-jp")
        self.assertEqual(test_message.context["user_profiles"][0]["speech"]
                         ["secondary_tts_language"], "it-it")

        # Change TTS language invalid
        test_message.data = {"utterance": "Change my text to speech language"}
        self.skill.handle_set_tts_language(test_message)
        self.skill.speak_dialog.assert_called_with("language_not_heard",
                                                   private=True)
        self.assertEqual(test_message.context["user_profiles"][0]["speech"]
                         ["tts_language"], "ja-jp")
        self.assertEqual(test_message.context["user_profiles"][0]["speech"]
                         ["secondary_tts_language"], "it-it")

        # Change TTS language unsupported
        test_message.data = {"utterance": "Change my text to speech language"
                                          "to bulgarian",
                             'rx_language': 'bulgarian'}
        self.skill.handle_set_tts_language(test_message)
        self.skill.speak_dialog.assert_called_with("language_not_supported",
                                                   {'io': 'speak',
                                                    'lang': 'Bulgarian'},
                                                   private=True)
        self.assertEqual(test_message.context["user_profiles"][0]["speech"]
                         ["tts_language"], "ja-jp")
        self.assertEqual(test_message.context["user_profiles"][0]["speech"]
                         ["secondary_tts_language"], "it-it")

        self.skill._languages = real_supported_languages

    def test_handle_set_language(self):
        real_set_stt_language = self.skill.handle_set_stt_language
        real_set_tts_language = self.skill.handle_set_tts_language
        self.skill.handle_set_stt_language = Mock()
        self.skill.handle_set_tts_language = Mock()

        # stt request
        stt_message = Message("test",
                              {"utterance": "set my language speech to text"})
        self.skill.handle_set_language(stt_message)
        self.skill.handle_set_stt_language.assert_called_once_with(stt_message)
        # tts request
        tts_message = Message("test",
                              {"utterance": "set my language text to speech"})
        self.skill.handle_set_language(tts_message)
        self.skill.handle_set_tts_language.assert_called_once_with(tts_message)

        # Unspecified request
        test_message = Message("test",
                               {"utterance": "set my language to spanish",
                                "rx_language": "spanish"})
        self.skill.handle_set_language(test_message)
        self.skill.handle_set_stt_language.assert_called_with(test_message)
        self.skill.handle_set_tts_language.assert_called_with(test_message)

        # Second language request
        second_lang_message = Message(
            "test", {"utterance": "change my secondary language to ukrainian",
                     "rx_language": "ukrainian", "second": "secondary"})
        self.skill.handle_set_language(second_lang_message)
        self.skill.handle_set_tts_language.assert_called_with(
            second_lang_message)

        # Unspecified STT not changed
        self.skill.handle_set_stt_language.reset_mock()
        test_profile = self.user_config
        test_profile["user"]["username"] = "test_user"
        test_profile["speech"]["stt_language"] = "en-us"
        test_message = Message("test", {"Language": "American English"},
                               {"username": "test_user",
                                "user_profiles": [test_profile]})
        self.skill.handle_set_language(test_message)
        self.skill.handle_set_tts_language.assert_called_with(test_message)
        self.skill.handle_set_stt_language.assert_not_called()

        self.skill.handle_set_stt_language = real_set_stt_language
        self.skill.handle_set_tts_language = real_set_tts_language

    def test_handle_no_secondary_language(self):
        test_profile = self.user_config
        test_profile["user"]["username"] = "test_user"
        test_profile["speech"]["secondary_tts_language"] = "es-es"
        test_message = Message("test", {},
                               {"username": "test_user",
                                "user_profiles": [test_profile]})
        self.skill.handle_no_secondary_language(test_message)
        self.skill.speak_dialog.assert_called_once_with("only_one_language",
                                                        private=True)
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["speech"]["secondary_tts_language"], "")
        self.assertEqual(test_message.context["user_profiles"][0]
                         ["speech"]["secondary_neon_voice"], "")
        profile = deepcopy(test_message.context["user_profiles"][0])

        self.skill.handle_no_secondary_language(test_message)
        self.skill.speak_dialog.assert_called_with("only_one_language",
                                                   private=True)
        self.assertEqual(test_message.context["user_profiles"][0], profile)

    def test_get_name_parts(self):
        user_profile = self.user_config
        # First none existing
        name = self.skill._get_name_parts("first", user_profile["user"])
        self.assertEqual(name, {"first_name": "first",
                                "full_name": "first"})

        # First and Last override First, existing Middle
        user_profile["user"]["first_name"] = "old"
        user_profile["user"]["middle_name"] = "middle"
        name = self.skill._get_name_parts("first last", user_profile["user"])
        self.assertEqual(name, {"first_name": "first",
                                "last_name": "last",
                                "full_name": "first middle last"})
        # First Middle Last override existing
        user_profile["user"]["first_name"] = "old"
        user_profile["user"]["middle_name"] = "old"
        user_profile["user"]["last_name"] = "old"
        name = self.skill._get_name_parts("first middle last",
                                          user_profile["user"])
        self.assertEqual(name, {"first_name": "first",
                                "middle_name": "middle",
                                "last_name": "last",
                                "full_name": "first middle last"})
        # Longer name override existing
        name = self.skill._get_name_parts("first-name middle last senior",
                                          user_profile["user"])
        self.assertEqual(name, {"first_name": "first-name",
                                "middle_name": "middle",
                                "last_name": "last senior",
                                "full_name": "first-name middle last senior"})

    def test_get_timezone_from_location(self):
        sleep(1)
        name, offset = \
            self.skill._get_timezone_from_location(
                self.skill._get_location_from_spoken_location("seattle"))
        self.assertEqual(name, "America/Los_Angeles")
        self.assertIsInstance(offset, float)

        timezone = \
            self.skill._get_timezone_from_location(None)
        self.assertIsNone(timezone)

    def test_get_location_from_spoken_location(self):
        # Test 'city' case
        sleep(1)
        address = self.skill._get_location_from_spoken_location("seattle")
        self.assertEqual(address['address']['city'], "Seattle")
        self.assertEqual(address['address']['state'], "Washington")
        self.assertEqual(address['address']['country'], "United States")
        self.assertIsInstance(address['lat'], str)
        self.assertIsInstance(address['lon'], str)

        # Test international case
        sleep(1)
        address = self.skill._get_location_from_spoken_location("kyiv",
                                                                "en-us")
        self.assertEqual(address['address']['city'], "Kyiv")
        self.assertEqual(address['address']['country'], "Ukraine")
        self.assertIsInstance(address['lat'], str)
        self.assertIsInstance(address['lon'], str)

        # Test 'town' case
        sleep(1)
        address = self.skill._get_location_from_spoken_location(
            "kirkland washington")
        self.assertEqual(address['address']['city'], "Kirkland")
        self.assertEqual(address['address']['state'], "Washington")
        self.assertEqual(address['address']['country'], "United States")
        self.assertIsInstance(address['lat'], str)
        self.assertIsInstance(address['lon'], str)

        # Test 'village' case
        sleep(1)
        address = self.skill._get_location_from_spoken_location(
            "orchard city colorado")
        self.assertEqual(address["address"]["city"], "Orchard City")
        self.assertEqual(address["address"]["state"], "Colorado")
        self.assertEqual(address['address']['country'], "United States")
        self.assertIsInstance(address['lat'], str)
        self.assertIsInstance(address['lon'], str)

    def test_parse_languages(self):
        self.assertEqual((None, None),
                         self.skill._parse_languages(
                             "change my language to french"))
        self.assertEqual(("australian english", None),
                         self.skill._parse_languages(
                             "change my primary language to australian "
                             "english"))
        self.assertEqual(("australian english", "mexican spanish"),
                         self.skill._parse_languages(
                             "change my primary language to australian "
                             "english and my second language to "
                             "mexican spanish"))

    def test_get_lang_name_and_code(self):
        from lingua_franca.internal import UnsupportedLanguageError
        # Check all the Coqui TTS Supported Languages
        self.assertEqual(("en-us", "American English"),
                         self.skill._get_lang_code_and_name("english"))
        self.assertEqual(("es-es", "Spanish"),
                         self.skill._get_lang_code_and_name("spanish"))
        self.assertEqual(("fr-fr", "French"),
                         self.skill._get_lang_code_and_name("french"))
        self.assertEqual(("de-de", "German"),
                         self.skill._get_lang_code_and_name("german"))
        self.assertEqual(("it-it", "Italian"),
                         self.skill._get_lang_code_and_name("italian"))
        self.assertEqual(("pl-pl", "Polish"),
                         self.skill._get_lang_code_and_name("polish"))
        self.assertEqual(("uk-ua", "Ukrainian"),
                         self.skill._get_lang_code_and_name("ukrainian"))
        self.assertEqual(("ro-ro", "Romanian"),
                         self.skill._get_lang_code_and_name("romanian"))
        self.assertEqual(("hu-hu", "Hungarian"),
                         self.skill._get_lang_code_and_name("hungarian"))
        self.assertEqual(("el-gr", "Greek"),
                         self.skill._get_lang_code_and_name("greek"))
        self.assertEqual(("sv-se", "Swedish"),
                         self.skill._get_lang_code_and_name("swedish"))
        self.assertEqual(("bg-bg", "Bulgarian"),
                         self.skill._get_lang_code_and_name("bulgarian"))
        self.assertEqual(("nl-nl", "Dutch"),
                         self.skill._get_lang_code_and_name("dutch"))
        self.assertEqual(("fi-fi", "Finnish"),
                         self.skill._get_lang_code_and_name("finnish"))
        self.assertEqual(("sl-si", "Slovenian"),
                         self.skill._get_lang_code_and_name("slovenian"))
        self.assertEqual(("lv-lv", "Latvian"),
                         self.skill._get_lang_code_and_name("latvian"))
        self.assertEqual(("et-ee", "Estonian"),
                         self.skill._get_lang_code_and_name("estonian"))
        self.assertEqual(("ga-ie", "Irish"),
                         self.skill._get_lang_code_and_name("irish"))

        # Check LibreTranslate supported languages
        self.assertEqual(("ar-sa", "Arabic"),
                         self.skill._get_lang_code_and_name("arabic"))
        self.assertEqual(("az-az", "Azerbaijani"),
                         self.skill._get_lang_code_and_name("azerbaijani"))
        self.assertEqual(("zh-zh", "Chinese"),
                         self.skill._get_lang_code_and_name("chinese"))
        self.assertEqual(("cs-cz", "Czech"),
                         self.skill._get_lang_code_and_name("czech"))
        self.assertEqual(("da-dk", "Danish"),
                         self.skill._get_lang_code_and_name("danish"))
        # self.assertEqual(("eo", "Esperanto"),
        #                  self.skill._get_lang_code_and_name("esperanto"))
        # self.assertEqual(("he-il", "Hebrew"),
        #                  self.skill._get_lang_code_and_name("hebrew"))
        self.assertEqual(("hi-in", "Hindi"),
                         self.skill._get_lang_code_and_name("hindi"))
        self.assertEqual(("id-id", "Indonesian"),
                         self.skill._get_lang_code_and_name("indonesian"))
        self.assertEqual(("ja-jp", "Japanese"),
                         self.skill._get_lang_code_and_name("japanese"))
        self.assertEqual(("ko-kr", "Korean"),
                         self.skill._get_lang_code_and_name("korean"))
        self.assertEqual(("fa-ir", "Persian"),
                         self.skill._get_lang_code_and_name("persian"))
        self.assertEqual(("pt-pt", "Portuguese"),
                         self.skill._get_lang_code_and_name("portuguese"))
        self.assertEqual(("ru-ru", "Russian"),
                         self.skill._get_lang_code_and_name("russian"))
        self.assertEqual(("sk-sk", "Slovak"),
                         self.skill._get_lang_code_and_name("slovak"))
        self.assertEqual(("tr-tr", "Turkish"),
                         self.skill._get_lang_code_and_name("turkish"))
        # Check manually specified alternative language requests
        self.assertEqual(("ga-ie", "Irish"),
                         self.skill._get_lang_code_and_name("gaelic"))
        self.assertEqual(("en-au", "English"),
                         self.skill._get_lang_code_and_name(
                             "australian english"))
        self.assertEqual(("fa-ir", "Persian"),
                         self.skill._get_lang_code_and_name("farsi"))
        with self.assertRaises(UnsupportedLanguageError):
            self.skill._get_lang_code_and_name("nothing")

    def test_spoken_email(self):
        self.assertEqual(self.skill._spoken_email("test@neon.ai"),
                         "test at neon dot ai")
        self.assertEqual(self.skill._spoken_email("my.email@domain.com"),
                         "my dot email at domain dot com")


if __name__ == '__main__':
    unittest.main()



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import re
from datetime import datetime
from threading import Event
from typing import Optional, Tuple
from adapt.intent import IntentBuilder
from dateutil.tz import gettz
from lingua_franca import load_language
from lingua_franca.time import default_timezone
from ovos_bus_client import Message
from neon_utils.location_utils import get_timezone
from neon_utils.skills.neon_skill import NeonSkill
from neon_utils.user_utils import get_user_prefs
from neon_utils.language_utils import get_supported_languages
from neon_utils.parse_utils import validate_email
from lingua_franca.parse import extract_langcode, get_full_lang_code
from lingua_franca.format import pronounce_lang
from lingua_franca.internal import UnsupportedLanguageError
from ovos_utils.file_utils import read_vocab_file
from ovos_utils import classproperty
from ovos_utils.log import LOG
from ovos_utils.process_utils import RuntimeRequirements
from mycroft.skills.core import intent_handler, intent_file_handler
from mycroft.util.parse import extract_datetime


class UserSettingsSkill(NeonSkill):
    MAX_SPEECH_SPEED = 1.5
    MIN_SPEECH_SPEED = 0.7

    def __init__(self, **kwargs):
        self._languages = None
        self._get_location = Event()
        NeonSkill.__init__(self, **kwargs)

    @classproperty
    def runtime_requirements(self):
        return RuntimeRequirements(network_before_load=False,
                                   internet_before_load=False,
                                   gui_before_load=False,
                                   requires_internet=True,
                                   requires_network=True,
                                   requires_gui=False,
                                   no_internet_fallback=True,
                                   no_network_fallback=True,
                                   no_gui_fallback=True)

    # TODO: move to __init__ after stable ovos-workshop release
    def initialize(self):
        if self.settings.get('use_geolocation'):
            LOG.debug(f"Geolocation update enabled")
            self.add_event("mycroft.ready", self._request_location_update)

    def _request_location_update(self, _=None):
        LOG.info(f'Requesting Geolocation update')
        self.add_event('ovos.ipgeo.update.response',
                       self._handle_location_ipgeo_update)
        self._get_location.clear()
        self.bus.emit(Message('ovos.ipgeo.update', {'overwrite': True}))
        if self._get_location.wait(10):
            return
        LOG.warning("No geolocation response, retrying...")
        self.bus.emit(Message('ovos.ipgeo.update', {'overwrite': True}))
        if not self._get_location.wait(30):
            LOG.error("No geolocation response")

    def _handle_location_ipgeo_update(self, message):
        self._get_location.set()
        updated_location = message.data.get('location')
        if not updated_location:
            LOG.warning(f"No geolocation returned by plugin")
            return
        from neon_utils.user_utils import apply_local_user_profile_updates
        from neon_utils.configuration_utils import NGIConfig
        user_config = NGIConfig("ngi_user_info")
        # default_coords = (
        #     str(self.config_core.default.get('location',
        #                                      {}).get('coordinate',
        #                                              {}).get('latitude')),
        #     str(self.config_core.default.get('location',
        #                                      {}).get('coordinate',
        #                                              {}).get('longitude')))
        # TODO: Figure out why default config doesn't match real default values
        default_coords = ("38.971669", "-95.23525")
        user_coords = (
            user_config.get('location', {}).get('lat'),
            user_config.get('location', {}).get('lng')
        )
        if not all(user_coords) or user_coords == default_coords:
            LOG.info(f'Updating default user config from ip geolocation')
            new_loc = {
                    'lat': str(updated_location['coordinate']['latitude']),
                    'lon': str(updated_location['coordinate']['longitude']),
                    'city': updated_location['city']['name'],
                    'state': updated_location['city']['state']['name'],
                    'country': updated_location['city']['state']['country']['name'],
                }
            name, offset = self._get_timezone_from_location(new_loc)
            new_loc['lng'] = new_loc.pop('lon')
            new_loc['tz'] = name
            new_loc['utc'] = str(round(offset, 1))
            apply_local_user_profile_updates({'location': new_loc}, user_config)
            self._emit_weather_update(message)
        else:
            LOG.debug(f'Ignoring IP location for already defined user location:'
                      f'{user_config["location"]}')
            LOG.debug(f'default={default_coords}')
            LOG.debug(f'user={user_coords}')
        # Remove listener after a successful update
        self.remove_event('ovos.ipgeo.update.response')

    @property
    def stt_languages(self) -> Optional[set]:
        self._get_supported_languages()
        if not all((self._languages.skills, self._languages.stt)):
            LOG.warning("Incomplete language support response. "
                        "Assuming all languages are supported")
            return None
        return set((lang for lang in self._languages.stt
                    if lang in self._languages.skills))

    @property
    def tts_languages(self):
        self._get_supported_languages()
        if not all((self._languages.skills, self._languages.tts)):
            LOG.warning("Incomplete language support response. "
                        "Assuming all languages are supported")
            return None
        return set((lang for lang in self._languages.tts
                    if lang in self._languages.skills))

    def _get_supported_languages(self):
        """
        Gather supported languages via the Messagebus API and save the result
        """
        if not self._languages:
            supported_langs = get_supported_languages()
            self._languages = supported_langs

    @intent_handler(IntentBuilder("ChangeUnits").require("change")
                    .require("units").one_of("imperial", "metric").build())
    def handle_unit_change(self, message: Message):
        """
        Handle a request to set metric or imperial units of measurement
        :param message: Message associated with request
        """
        new_unit = "imperial" if message.data.get("imperial") else \
            "metric" if message.data.get("metric") else None
        if not new_unit:
            raise RuntimeError("Missing required imperial or metric vocab")

        current_unit = get_user_prefs(message)["units"]["measure"]
        if new_unit == current_unit:
            self.speak_dialog("units_already_set", {"unit": new_unit},
                              private=True)
        else:
            updated_prefs = {"units": {"measure": new_unit}}
            self.update_profile(updated_prefs, message)
            self.speak_dialog("units_changed",
                              {"unit": self.translate(f"word_{new_unit}")},
                              private=True)
            self._emit_weather_update(message)

    @intent_handler(IntentBuilder("ChangeTime").require("change")
                    .require("time").one_of("half", "full").build())
    def handle_time_format_change(self, message: Message):
        """
        Handle a request to set time format to 12 or 24 hour time
        :param message: Message associated with request
        """
        new_setting = 12 if message.data.get("half") else \
            24 if message.data.get("full") else None
        if not new_setting:
            raise RuntimeError("Missing required time scale vocab")

        current_setting = get_user_prefs(message)["units"]["time"]
        if new_setting == current_setting:
            self.speak_dialog("time_format_already_set",
                              {"scale": str(new_setting)}, private=True)
        else:
            updated_prefs = {"units": {"time": new_setting}}
            self.update_profile(updated_prefs, message)
            self.speak_dialog("time_format_changed",
                              {"scale": str(new_setting)}, private=True)

    @intent_handler(IntentBuilder("ChangeDate").require("change")
                    .require("date").one_of("mdy", "dmy", "ymd").build())
    def handle_date_format_change(self, message: Message):
        """
        Handle a request to set date format to DMY, MDY, or YMD format
        :param message: Message associated with request
        """
        new_setting = "YMD" if message.data.get("ymd") else \
            "MDY" if message.data.get("mdy") else \
            "DMY" if message.data.get("dmy") else None
        if not new_setting:
            raise RuntimeError("Missing required date format vocab")

        current_setting = get_user_prefs(message)["units"]["date"]
        if new_setting == current_setting:
            self.speak_dialog("date_format_already_set",
                              {"format": message.data.get(new_setting.lower())},
                              private=True)
        else:
            updated_prefs = {"units": {"date": new_setting}}
            self.update_profile(updated_prefs, message)
            self.speak_dialog("date_format_changed",
                              {"format": message.data.get(new_setting.lower())},
                              private=True)

    @intent_handler(IntentBuilder("SetHesitation").one_of("permit", "deny")
                    .require("hesitation").build())
    def handle_speak_hesitation(self, message: Message):
        """
        Handle a request for Neon to speak something when intent processing
        may take some time
        :param message: Message associated with request
        """
        enabled = True if message.data.get("permit") else False
        self.update_profile({"response_mode": {"hesitation": enabled}})
        if enabled:
            self.speak_dialog("hesitation_enabled", private=True)
        else:
            self.speak_dialog("hesitation_disabled", private=True)

    @intent_handler(IntentBuilder("Transcription").one_of("permit", "deny")
                    .optionally("audio").optionally("text").require("retention")
                    .build())
    def handle_transcription_retention(self, message: Message):
        """
        Handle a request to permit or deny saving audio recordings
        :param message: Message associated with request
        """
        allow = True if message.data.get("permit") else False
        kind = "save_audio" if message.data.get("audio") else \
            "save_text" if message.data.get("text") else None
        if not kind:
            LOG.warning(f"No transcription type specified, assume text")
            kind = 'save_text'

        transcription = "word_audio" if kind == "save_audio" else "word_text"
        enabled = "word_enabled" if allow else "word_disabled"

        current_setting = get_user_prefs(message)["privacy"][kind]
        if current_setting == allow:
            self.speak_dialog("transcription_already_set",
                              {"transcription": self.translate(transcription),
                               "enabled": self.translate(enabled)},
                              private=True)
        else:
            updated_prefs = {"privacy": {kind: allow}}
            self.update_profile(updated_prefs, message)
            self.speak_dialog("transcription_changed",
                              {"transcription": self.translate(transcription),
                               "enabled": self.translate(enabled)},
                              private=True)

    @intent_handler(IntentBuilder("SpeakSpeed").require("speak_to_me")
                    .one_of("faster", "slower", "normally").build())
    def handle_speech_speed(self, message: Message):
        """
        Handle a request to adjust response audio playback speed
        :param message: Message associated with request
        """
        current_speed = float(get_user_prefs(message)["speech"].get(
            "speed_multiplier")) or 1.0
        if message.data.get("faster"):
            speed = current_speed / 0.9
        elif message.data.get("slower"):
            speed = current_speed * 0.9
        elif message.data.get("normally"):
            speed = 1.0
        else:
            raise RuntimeError("Missing speed keyword")

        if speed < self.MIN_SPEECH_SPEED:
            speed = self.MIN_SPEECH_SPEED
        elif speed > self.MAX_SPEECH_SPEED:
            speed = self.MAX_SPEECH_SPEED

        speed = round(speed, 1)
        self.update_profile({"speech": {"speed_multiplier": speed}})

        if speed == current_speed == self.MAX_SPEECH_SPEED:
            self.speak_dialog("speech_speed_limit",
                              {"limit": self.translate("word_faster")},
                              private=True)
        elif speed == current_speed == self.MIN_SPEECH_SPEED:
            self.speak_dialog("speech_speed_limit",
                              {"limit": self.translate("word_slower")},
                              private=True)
        elif speed == 1.0:
            self.speak_dialog("speech_speed_normal", private=True)
        elif speed > current_speed:
            self.speak_dialog("speech_speed_faster", private=True)
        elif speed < current_speed:
            self.speak_dialog("speech_speed_slower", private=True)

    @intent_handler(IntentBuilder("ChangeLocationTimezone").require("change")
                    .one_of("timezone", "location").require("rx_place")
                    .build())
    def handle_change_location_timezone(self, message: Message):
        """
        Handle a request to change user configured location or timezone.
        This will prompt the user to update the non-requested setting too
        :param message: Message associated with request
        """
        requested_place = message.data.get("rx_place")
        resolved_place = self._get_location_from_spoken_location(
            requested_place, self.lang)
        if not resolved_place and message.data.get("timezone"):
            # TODO: Try resolving tz by name DM
            pass
        if not resolved_place:
            self.speak_dialog("location_not_found",
                              {"location": requested_place},
                              private=True)
            return

        tz_name, utc_offset = get_timezone(resolved_place["lat"],
                                           resolved_place["lon"])
        if message.data.get("timezone"):
            do_timezone = True
            do_location = self.ask_yesno(
                "also_change_location_tz",
                {"type": self.translate("word_location"),
                 "new": requested_place}) == "yes"
        elif message.data.get("location"):
            do_location = True
            do_timezone = self.ask_yesno(
                "also_change_location_tz",
                {"type": self.translate("word_timezone"),
                 "new": requested_place}) == "yes"
        else:
            do_location = False
            do_timezone = False

        if do_timezone:
            self.update_profile({"location": {"tz": tz_name,
                                              "utc": utc_offset}})
            self.speak_dialog("change_location_tz",
                              {"type": self.translate("word_timezone"),
                               "location": f"UTC {utc_offset}"},
                              private=True)
        if do_location:
            LOG.info(f"Update location: {resolved_place}")
            self.update_profile({'location': {
                'city': resolved_place['address']['city'],
                'state': resolved_place['address'].get('state'),
                'country': resolved_place['address']['country'],
                'lat': float(resolved_place['lat']),
                'lng': float(resolved_place['lon'])}})
            self.speak_dialog("change_location_tz",
                              {"type": self.translate("word_location"),
                               "location": resolved_place['address']['city']},
                              private=True)
            self._emit_weather_update(message)

    @intent_handler(IntentBuilder("ChangeDialog").one_of("change", "permit")
                    .require("dialog_mode").one_of("random", "limited")
                    .build())
    def handle_change_dialog_mode(self, message: Message):
        """
        Handle a request to switch between normal and limited dialog modes
        :param message: Message associated with request
        """
        new_dialog = "word_random" if message.data.get("random") else \
            "word_limited" if message.data.get("limited") else None
        if not new_dialog:
            raise RuntimeError("Missing required dialog mode")
        new_limit_dialog = new_dialog == "word_limited"
        current_limit_dialog = get_user_prefs(message)["response_mode"].get(
            "limit_dialog", False)

        if new_limit_dialog == current_limit_dialog:
            self.speak_dialog("dialog_mode_already_set",
                              {"response": self.translate(new_dialog)},
                              private=True)
            return

        self.update_profile(
            {"response_mode": {"limit_dialog": new_limit_dialog}})
        self.speak_dialog("dialog_mode_changed",
                          {"response": self.translate(new_dialog)},
                          private=True)

    @intent_handler(IntentBuilder("SayMyName").require("tell_me_my")
                    .require("name").build())
    @intent_file_handler("who_am_i.intent")
    def handle_say_my_name(self, message: Message):
        """
        Handle a request to read back a user's name
        :param message: Message associated with request
        """
        if not self.neon_in_request(message):
            return
        utterance = message.data.get("utterance")
        profile = get_user_prefs(message)
        # a 32-char username is a generated UID, not a real username
        real_username = len(profile["user"]["username"]) < 32
        if not any((profile["user"]["first_name"],
                    profile["user"]["middle_name"],
                    profile["user"]["last_name"],
                    profile["user"]["preferred_name"],
                    profile["user"]["full_name"])) and not real_username:
            # TODO: Use get_response to ask for the user's name
            self.speak_dialog(
                "name_not_known",
                {"name_position": self.translate("word_name")},
                private=True)
            return
        if self.voc_match(utterance, "first_name"):
            name = profile["user"]["first_name"]
            request = "word_first_name"
        elif self.voc_match(utterance, "middle_name"):
            name = profile["user"]["middle_name"]
            request = "word_middle_name"
        elif self.voc_match(utterance, "last_name"):
            name = profile["user"]["last_name"]
            request = "word_last_name"
        elif self.voc_match(utterance, "preferred_name"):
            name = profile["user"]["preferred_name"]
            request = "word_preferred_name"
        elif self.voc_match(utterance, "full_name"):
            name = profile["user"]["full_name"]
            request = "word_full_name"
        elif self.voc_match(utterance, "username"):
            name = profile["user"]["username"]
            request = "word_username"
        else:
            name = profile["user"]["preferred_name"] or \
                   profile["user"]["first_name"] or \
                   profile["user"]["username"]
            request = "word_name"

        if not name or name == profile["user"]["username"] == 'local':
            if request == "word_username":
                self.speak_dialog("name_no_username", private=True)
            else:
                # TODO: Use get_response to ask for the user's name
                self.speak_dialog("name_not_known",
                                  {"name_position": self.translate(request)},
                                  private=True)
        else:
            self.speak_dialog("name_is",
                              {"name_position": self.translate(request),
                               "name": name}, private=True)

    @intent_handler(IntentBuilder("SayMyEmail").require("tell_me_my")
                    .require("email").build())
    def handle_say_my_email(self, message: Message):
        """
        Handle a request to read back the user's email address
        :param message: Message associated with request
        """
        if not self.neon_in_request(message):
            return
        email_address = get_user_prefs(message)["user"]["email"]
        if not email_address:
            # TODO: Use get_response to ask for the user's email
            self.speak_dialog("email_not_known", private=True)
        else:
            self.speak_dialog("email_is",
                              {"email": self._spoken_email(email_address)},
                              private=True)

    @intent_handler(IntentBuilder("SayMyLocation").require("tell_me_my")
                    .require("location").build())
    @intent_file_handler("where_am_i.intent")
    def handle_say_my_location(self, message: Message):
        """
        Handle a request to read back the user's location
        :param message: Message associated with request
        """
        if not self.neon_in_request(message):
            return
        location_prefs = get_user_prefs(message)["location"]
        if not location_prefs["city"]:
            from neon_utils.net_utils import check_online
            if check_online():
                self.speak_dialog("location_unknown_online",
                                  private=True)
            else:
                self.speak_dialog("location_unknown_offline",
                                  private=True)
        else:
            friendly_location = ", ".join([x for x in
                                           (location_prefs["city"],
                                            location_prefs["state"] or
                                            location_prefs["country"])])
            self.speak_dialog("location_is", {"location": friendly_location},
                              private=True)

    @intent_handler(IntentBuilder("SayMyBirthday").require("tell_me_my")
                    .require("birthday").build())
    @intent_file_handler("when_is_my_birthday.intent")
    def handle_say_my_birthday(self, message: Message):
        """
        Handle a request to read back the user's birthday
        :param message: Message associated with request
        """
        if not self.neon_in_request(message):
            return
        birthday_str = get_user_prefs(message)["user"]["dob"]
        if not birthday_str or birthday_str == "YYYY/MM/DD":
            self.speak_dialog("birthday_not_known", private=True)
            return
        user_tz = gettz(self.location_timezone) if self.location_timezone else \
            default_timezone()
        now_time = datetime.now(user_tz)
        birthday_dt = datetime.strptime(birthday_str, "%Y/%m/%d")
        speakable_birthday = birthday_dt.strftime("%B %-d")
        self.speak_dialog("birthday_is",
                          {"birthday": speakable_birthday}, private=True)

        if birthday_dt.month == now_time.month and \
                birthday_dt.day == now_time.day:
            self.speak_dialog("happy_birthday", private=True)

    @intent_handler(IntentBuilder("SetMyBirthday").require("my")
                    .require("birthday").build())
    def handle_set_my_birthday(self, message: Message):
        """
        Handle a request to set a user's birthday
        :param message: Message associated with request
        """
        if not self.neon_in_request(message):
            return
        load_language(self.lang)
        user_tz = gettz(self.location_timezone) if self.location_timezone else \
            default_timezone()
        now_time = datetime.now(user_tz)
        try:
            birth_date, _ = extract_datetime(message.data.get("utterance"),
                                             now_time, self.lang)
        except IndexError:
            self.speak_dialog("birthday_not_heard", private=True)
            return

        formatted_birthday = birth_date.strftime("%Y/%m/%d")
        # TODO: Update to use LF when method added for month + date format
        # anchor_date = now_time.replace(year=birth_date.year)
        # speakable_birthday = nice_date(birth_date, now=anchor_date)
        speakable_birthday = birth_date.strftime("%B %-d")

        self.update_profile({"user": {"dob": formatted_birthday}}, message)
        self.speak_dialog("birthday_confirmed",
                          {"birthday": speakable_birthday}, private=True)

        if birth_date.month == now_time.month and \
                birth_date.day == now_time.day:
            self.speak_dialog("happy_birthday", private=True)

    @intent_handler(IntentBuilder("SetMyEmail").optionally("change")
                    .require("my").require("email").require("rx_setting")
                    .build())
    def handle_set_my_email(self, message: Message):
        """
       Handle a request to set a user's email address
       :param message: Message associated with request
       """
        # Parse actual email address from intent
        extracted = message.data.get("rx_setting")
        LOG.debug(extracted)
        email_addr: str = extracted.split()[0] + \
            message.data.get("utterance").rsplit(extracted.split()[0])[1]
        dot = read_vocab_file(self.find_resource("dot.voc", 'vocab'))[0][0]
        at = read_vocab_file(self.find_resource("at.voc", 'vocab'))[0][0]
        email_words = email_addr.split()
        if dot in email_words:
            email_words[email_words.index(dot)] = "."
        if at in email_words:
            email_words[email_words.index(at)] = "@"
        email_addr = "".join(email_words)
        LOG.info(email_addr)

        if not validate_email(email_addr):
            self.speak_dialog("email_set_error", private=True)
            email_addr = self.get_gui_input(self.translate("word_email_title"),
                                            "test@neon.ai")
            if not email_addr or not validate_email(email_addr):
                LOG.warning(f"Invalid email_addr entered: {email_addr}")
                return

        current_email = get_user_prefs(message)["user"]["email"]
        if current_email and email_addr == current_email:
            self.speak_dialog("email_already_set_same",
                              {"email": self._spoken_email(current_email)},
                              private=True)
            return
        if current_email:
            if self.ask_yesno("email_overwrite",
                              {"old": self._spoken_email(current_email),
                               "new": self._spoken_email(email_addr)}) == "yes":
                self.update_profile({"user": {"email": email_addr}})
                self.speak_dialog("email_set",
                                  {"email": self._spoken_email(email_addr)},
                                  private=True)
            else:
                self.speak_dialog("email_not_changed",
                                  {"email": self._spoken_email(current_email)},
                                  private=True)
            return
        if self.ask_yesno("email_confirmation",
                          {"email": self._spoken_email(email_addr)}) == "yes":
            self.update_profile({"user": {"email": email_addr}})
            self.speak_dialog("email_set",
                              {"email": self._spoken_email(email_addr)},
                              private=True)
        else:
            self.speak_dialog("email_not_confirmed", private=True)
            email_addr = self.get_gui_input(self.translate("word_email_title"),
                                            "test@neon.ai")
            if email_addr:
                self.update_profile({"user": {"email": email_addr}})
                self.speak_dialog("email_set",
                                  {"email": self._spoken_email(email_addr)},
                                  private=True)
            else:
                LOG.info("User Cancelled email input")
                # TODO: Speak confirmation

    def get_gui_input(self, title=None, placeholder=None,
                      confirm_text=None, exit_text=None) -> Optional[str]:
        gui_response = None
        response_event = Event()
        response_event.clear()
        self.gui.show_input_box(title, placeholder, confirm_text, exit_text,
                                True, True)

        def _on_response(message):
            nonlocal gui_response
            gui_response = message.data.get("text")
            response_event.set()

        def _on_close(message):
            response_event.set()

        resp_message = self.gui.build_message_type('input.box.response')
        close_message = self.gui.build_message_type('input.box.close')
        self.add_event(resp_message, _on_response, once=True)
        self.add_event(close_message, _on_close, once=True)
        response_event.wait()
        self.gui.remove_input_box()
        self.remove_event(resp_message)
        self.remove_event(close_message)
        return gui_response

    @intent_handler(IntentBuilder("SetMyName").optionally("change")
                    .require("my").require("name").require("rx_setting")
                    .build())
    @intent_handler(IntentBuilder("MyNameIs").require("my_name_is")
                    .require("rx_name").build())
    def handle_set_my_name(self, message: Message):
        """
        Handle a request to set a user's name. Some considerations for name
        parsing can be found from w3c:
        https://www.w3.org/International/questions/qa-personal-names
        :param message: Message associated with request
        """
        if not self.neon_in_request(message):
            return
        confirmed = True
        utterance = message.data.get("utterance")
        name = message.data.get("rx_setting") or message.data.get("rx_name")
        if self.voc_match(utterance, "first_name"):
            request = "first_name"
            name = name.title()
        elif self.voc_match(utterance, "middle_name"):
            name = name.title()
            request = "middle_name"
        elif self.voc_match(utterance, "last_name"):
            name = name.title()
            request = "last_name"
        elif self.voc_match(utterance, "preferred_name"):
            name = name.title()
            request = "preferred_name"
        elif self.voc_match(utterance, "username"):
            self.speak_dialog("error_change_username", private=True)
            return
        else:
            name = name.title()
            request = None

        user_profile = get_user_prefs(message)["user"]

        # Catch an invalid intent match
        if (request and len(name.split()) > 3) or len(name.split()) > 4:
            LOG.warning(f"'{name}' does not look like a {request} name.")
            confirmed = self.ask_yesno("name_confirm_change",
                                       {"position": self.translate(
                                           f"word_{request or 'name'}"),
                                        "name": name})

        if not confirmed:
            self.speak_dialog("name_not_confirmed", private=True)
        elif request:
            if name == user_profile[request]:
                self.speak_dialog(
                    "name_not_changed",
                    {"position": self.translate(f"word_{request}"),
                     "name": name}, private=True)
            elif confirmed:
                name_parts = (name if request == n else user_profile.get(n)
                              for n in ("first_name", "middle_name",
                                        "last_name"))
                full_name = " ".join((n for n in name_parts if n))
                self.update_profile({"user": {request: name,
                                              "full_name": full_name}},
                                    message)
                self.speak_dialog(
                    "name_set_part",
                    {"position": self.translate(f"word_{request}"),
                     "name": name}, private=True)
        else:
            preferred_name = user_profile["preferred_name"] or name
            name_parts = self._get_name_parts(name, user_profile)
            if preferred_name == user_profile["first_name"] and \
                    "first_name" in name_parts:
                preferred_name = name_parts["first_name"]
            updated_user_profile = {"preferred_name": preferred_name,
                                    **name_parts}
            if all((user_profile[n] == updated_user_profile.get(n) for n in
                    ("first_name", "middle_name", "last_name"))):
                self.speak_dialog("name_not_changed",
                                  {"position": self.translate(f"word_name"),
                                   "name": name})
            else:
                self.update_profile({"user": updated_user_profile}, message)
                self.speak_dialog("name_set_full",
                                  {"nick": preferred_name,
                                   "name": name_parts["full_name"]},
                                  private=True)

    @intent_handler(IntentBuilder("SayMyLanguageSettings")
                    .require("tell_me_my").require("language_settings")
                    .build())
    @intent_file_handler("language_settings.intent")
    def handle_say_my_language_settings(self, message: Message):
        """
        Handle a request to read back the user's language settings
        :param message: Message associated with request
        """
        load_language(self.lang)
        language_settings = get_user_prefs(message)["speech"]
        primary_lang = pronounce_lang(language_settings["tts_language"])
        second_lang = pronounce_lang(
            language_settings["secondary_tts_language"])
        self.speak_dialog("language_setting",
                          {"primary": self.translate("word_primary"),
                           "language": primary_lang,
                           "gender": self.translate(
                               f'word_{language_settings["tts_gender"]}')},
                          private=True)
        if second_lang and (second_lang != primary_lang or
                            language_settings["tts_gender"] !=
                            language_settings["secondary_tts_gender"]):
            self.speak_dialog(
                "language_setting",
                {"primary": self.translate("word_secondary"),
                 "language": second_lang,
                 "gender": self.translate(
                     f'word_{language_settings["secondary_tts_gender"]}')},
                private=True)

    @intent_handler(IntentBuilder("SetSTTLanguage").require("change")
                    .optionally("my").require("language_stt")
                    .require("language").require("rx_language").build())
    @intent_file_handler("language_stt.intent")
    def handle_set_stt_language(self, message: Message):
        """
        Handle a request to change the language spoken by the user
        :param message: Message associated with request
        """
        requested_lang = message.data.get('rx_language') or \
            message.data.get('request_language')
        lang = self._parse_languages(message.data.get("utterance"))[0] or \
            requested_lang.split()[-1]
        try:
            code, spoken_lang = self._get_lang_code_and_name(lang)
        except UnsupportedLanguageError as e:
            LOG.error(e)
            self.speak_dialog("language_not_recognized", {"lang": lang},
                              private=True)
            return

        LOG.info(f"code={code}")
        if self.stt_languages and code.split('-')[0] not in self.stt_languages:
            LOG.warning(f"{code} not found in: {self.stt_languages}")
            self.speak_dialog("language_not_supported",
                              {"lang": spoken_lang,
                               "io": self.translate('word_understand')},
                              private=True)
            return
        dialog_data = {"io": self.translate("word_stt"),
                       "lang": spoken_lang}
        if code == get_user_prefs(message)["speech"]["stt_language"]:
            self.speak_dialog("language_not_changed", dialog_data,
                              private=True)
            return

        if self.ask_yesno("language_change_confirmation",
                          dialog_data) == "yes":
            self.update_profile({"speech": {"stt_language": code}})
            self.speak_dialog("language_set", dialog_data,
                              private=True)
        else:
            self.speak_dialog("language_not_confirmed", private=True)

    @intent_handler(IntentBuilder("SetTTSLanguage").require("change")
                    .optionally("my").require("language_tts")
                    .require("language").require("rx_language").build())
    @intent_file_handler("language_tts.intent")
    def handle_set_tts_language(self, message: Message):
        """
        Handle a request to change the language spoken to the user
        :param message: Message associated with request
        """
        language = message.data.get("rx_language") or \
            message.data.get("request_language")
        primary, secondary = \
            self._parse_languages(message.data.get("utterance"))
        LOG.info(f"primary={primary} | secondary={secondary} | "
                 f"language={language}")
        user_settings = get_user_prefs(message)
        if primary:
            try:
                primary_code, primary_spoken = \
                    self._get_lang_code_and_name(primary)
                LOG.info(f"primary={primary_code}")
                if self.tts_languages and \
                        primary_code.split('-')[0] not in self.tts_languages:
                    LOG.warning(f"{primary_code} not found in:"
                                f" {self.tts_languages}")
                    self.speak_dialog("language_not_supported",
                                      {"lang": primary_spoken,
                                       "io": self.translate('word_speak')},
                                      private=True)
                    return
                gender = self._get_gender(primary) or \
                    user_settings["speech"]["tts_gender"]
                self.update_profile({"speech": {"tts_gender": gender,
                                                "tts_language": primary_code}},
                                    message)
                self.speak_dialog("language_set",
                                  {"io": self.translate("word_primary"),
                                   "lang": primary_spoken}, private=True)
            except UnsupportedLanguageError:
                LOG.warning(f"No language for primary request: {primary}")
                self.speak_dialog("language_not_recognized", {"lang": primary},
                                  private=True)
        if secondary:
            try:
                secondary_code, secondary_spoken = \
                    self._get_lang_code_and_name(secondary)
                LOG.info(f"secondary={secondary_code}")
                if self.tts_languages and \
                        secondary_code.split('-')[0] not in self.tts_languages:
                    LOG.warning(f"{secondary_code} not found in:"
                                f" {self.tts_languages}")
                    self.speak_dialog("language_not_supported",
                                      {"lang": secondary_spoken,
                                       "io": self.translate('word_speak')},
                                      private=True)
                    return
                gender = self._get_gender(secondary) or \
                    user_settings["speech"]["secondary_tts_gender"]
                self.update_profile(
                    {"speech": {"secondary_tts_gender": gender,
                                "secondary_tts_language": secondary_code}},
                    message)
                self.speak_dialog("language_set",
                                  {"io": self.translate("word_secondary"),
                                   "lang": secondary_spoken}, private=True)
            except UnsupportedLanguageError:
                LOG.warning(f"No language for secondary request: {secondary}")
                self.speak_dialog("language_not_recognized",
                                  {"lang": secondary}, private=True)

        if any((primary, secondary)):
            return
        if language:
            try:
                code, spoken = \
                    self._get_lang_code_and_name(language)
                if self.tts_languages and \
                        code.split('-')[0] not in self.tts_languages:
                    LOG.warning(f"{code} not found in: {self.tts_languages}")
                    self.speak_dialog("language_not_supported",
                                      {"lang": spoken,
                                       "io": self.translate('word_speak')},
                                      private=True)
                    return
                gender = self._get_gender(language) or \
                    user_settings["speech"]["tts_gender"]
                self.update_profile({"speech": {"tts_gender": gender,
                                                "tts_language": code}},
                                    message)
                self.speak_dialog("language_set",
                                  {"io": self.translate("word_primary"),
                                   "lang": spoken}, private=True)
            except UnsupportedLanguageError:
                LOG.warning(f"No language for secondary request: {language}")
                self.speak_dialog("language_not_recognized",
                                  {"lang": language}, private=True)
        else:
            LOG.warning("No language parsed")
            self.speak_dialog("language_not_heard", private=True)

    @intent_handler(IntentBuilder("SetMyLanguage").optionally("change")
                    .require("my").optionally("preferred").optionally("second")
                    .require("language").optionally("rx_language").build())
    def handle_set_language(self, message: Message):
        """
        Handle a user request to change languages. Checks for improper parsing
        of STT/TTS keywords and otherwise updates both STT and TTS settings.
        :param message: Message associated with request
        """
        utterance = message.data.get("utterance")
        LOG.info(f"language={message.data.get('language')}")
        LOG.info(f"preferred={message.data.get('preferred')}|"
                 f"second={message.data.get('second')}")
        LOG.info(f"Ambiguous language change request: {utterance}")
        if self.voc_match(utterance, "language_stt"):
            LOG.warning("STT Intent not matched")
            self.handle_set_stt_language(message)
        elif self.voc_match(utterance, "language_tts"):
            LOG.warning("TTS Intent not matched")
            self.handle_set_tts_language(message)
        elif message.data.get('second'):
            # TODO: This should also set second STT language if applicable
            self.handle_set_tts_language(message)
        else:
            LOG.info("General language change request, handle STT+TTS")
            self.handle_set_tts_language(message)
            try:
                lang = self._get_lang_code_and_name(
                    message.data.get("rx_language", ""))[0]
                if not lang or lang != \
                        get_user_prefs(message)["speech"]["stt_language"]:
                    self.handle_set_stt_language(message)
            except UnsupportedLanguageError:
                pass

    @intent_handler(IntentBuilder("NoSecondaryLanguage")
                    .require("no_secondary_language").build())
    def handle_no_secondary_language(self, message: Message):
        """
        Handle a user request to only hear responses in one language
        :param message: Message associated with request
        """
        self.update_profile({"speech": {"secondary_tts_language": "",
                                        "secondary_neon_voice": ""}},
                            message)
        self.speak_dialog("only_one_language", private=True)

    def _emit_weather_update(self, message: Message):
        """
        Emit a weather update on location change
        """
        self.bus.emit(
            message.forward("skill-ovos-weather.openvoiceos.weather.request"))

    def _parse_languages(self, utterance: str) -> \
            (Optional[str], Optional[str]):
        """
        Parse a language change request for primary and secondary languages
        :param utterance: raw utterance spoken by the user
        :returns: spoken primary, secondary languages requested
        """
        def _get_rx_patterns(rx_file: str, utt: str):
            with open(rx_file) as f:
                for pat in f.read().splitlines():
                    pat = pat.strip()
                    if pat and pat[0] == "#":
                        continue
                    res = re.search(pat, utt)
                    if res:
                        return res
        utterance = f"{utterance}\n"
        primary_tts = self.find_resource('primary_tts.rx', 'regex')
        secondary_tts = self.find_resource('secondary_tts.rx', 'regex')
        if primary_tts:
            try:
                primary = _get_rx_patterns(primary_tts,
                                           utterance).group("rx_primary").strip()
            except (IndexError, AttributeError):
                primary = None
        else:
            LOG.warning("Could not resolve primary_tts.rx")
            primary = None
        if secondary_tts:
            try:
                secondary = _get_rx_patterns(secondary_tts,
                                             utterance).group("rx_secondary")\
                    .strip()
            except (IndexError, AttributeError):
                secondary = None
        else:
            LOG.warning("Could not resolve secondary_tts.rx")
            secondary = None

        return primary, secondary

    def _get_lang_code_and_name(self, request: str) -> (str, str):
        """
        Extract the lang code and pronounceable name from a requested language
        :param request: user requested language
        :returns: lang code and pronounceable language name if found, else None
        """
        load_language(self.lang)

        code = None
        # Manually specified languages take priority
        request_overrides = self.translate_namedvalues("languages.value")
        for lang, c in request_overrides.items():
            if lang in request.lower().split():
                code = c
                break
        if not code:
            # Ask LF to determine the code
            short_code = extract_langcode(request)[0]
            code = get_full_lang_code(short_code)
            if code.split('-')[0] != short_code:
                LOG.warning(f"Got {code} from {short_code}. No valid code")
                code = None

        if not code:
            # Request is not a language, raise an exception
            raise UnsupportedLanguageError(f"No language found in {request}")
        spoken_lang = pronounce_lang(code)
        return code, spoken_lang

    def _get_gender(self, request: str) -> Optional[str]:
        """
        Extract a requested voice gender
        :param request: Parsed user requested language
        :returns: 'male', 'female', or None
        """
        if self.voc_match(request, "male"):
            return "male"
        if self.voc_match(request, "female"):
            return "female"
        LOG.info(f"no gender in request: {request}")
        return None

    def _spoken_email(self, email_addr: str):
        """
        Get a pronouncable email address string
        """
        return email_addr.replace('.', f' {self.translate("word_dot")} ')\
            .replace('@', f' {self.translate("word_at")} ')

    @staticmethod
    def _get_name_parts(name: str, user_profile: dict) -> dict:
        """
        Parse a name string into first/middle/last components
        :param user_profile: user preferences dict with keys:
            ('first_name', 'middle_name', 'last_name')
        :returns: dict of positional names extracted
        """
        name_parts = name.split()
        if len(name_parts) == 1:
            name = {"first_name": name}
        elif len(name_parts) == 2:
            name = {"first_name": name_parts[0],
                    "last_name": name_parts[1]}
        elif len(name_parts) == 3:
            name = {"first_name": name_parts[0],
                    "middle_name": name_parts[1],
                    "last_name": name_parts[2]}
        else:
            LOG.warning(f"Longer name than expected: {name}")
            # TODO: Better logic here
            name = {"first_name": name_parts[0],
                    "middle_name": name_parts[1],
                    "last_name": " ".join(name_parts[2:])}
        name_parts = (name.get(n) or user_profile.get(n)
                      for n in ("first_name", "middle_name", "last_name"))
        name["full_name"] = " ".join((n for n in name_parts if n))
        return name

    @staticmethod
    def _get_timezone_from_location(location: dict) -> \
            Optional[Tuple[str, float]]:
        """
        Get timezone info for the resolved location
        :param location: location data to get timezone for
        :returns: Timezone name, UTC Offset
        """
        try:
            tz_name, tz_offset = get_timezone(location["lat"], location["lon"])
            return tz_name, tz_offset
        except (KeyError, TypeError):
            return None

    @staticmethod
    def _get_location_from_spoken_location(location: str,
                                           lang: Optional[str] = None) -> \
            Optional[dict]:
        """
        Get address information for the requested location
        :param location: spoken location to get data for
        :returns: dict of location data containing at minimum:
            'lat', 'lon', 'address'['city'], address['country']
        """
        from neon_utils.location_utils import get_full_location
        try:
            place = get_full_location(location, lang)
            if not place or not place.get("address"):
                LOG.warning(f"Could not locate: {location}")
                return None
            if not place['address'].get("city"):
                place["address"]["city"] = place["address"].get("town") or \
                                           place["address"].get("village")
        except AttributeError:
            LOG.warning(f"Could not locate: {location}")
            return None
        return place

    def stop(self):
        pass
